#eyam
# Model of the Black Death plague outbreak in the English village of Eyam
# using data recorded by Mompesson in 1666.
#
# Times are in months ( 1 month is defined as 365/12 days )
# I mean population of Eyam that are Infected
# S means population of Eyam that are Succeptable to infection. In this model
# it is assumed there is no natural immunity.
# D means population that have died since t=0
# N = S + I is the population at time t
# N0 = S + I + D is the population at time t=0
#
# Based upon a Div project by Dr John Cullerne, Winchester College.
#
# LAST UPDATED by Andy French December 2017

#Import Python numeric, math and plotting libraries
import numpy as Npy
import math
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

#

#Model and plotting input parameters

#Define timestep
dt = 0.1

#Define end t value /months
tmax = 4

#Define range and number of alpha iterations
alphamin = 2.7
alphamax = 3.1
num_alpha = 100

#Fontsize and markersize for plots
fsize = 22
msize = 14

#

#Euler method solver for differential equations which
#describe model of Eyam epidemic. Note D[0] = 0
def eyam_model( dt, I0, S0, alpha, beta, tmax ):

    #Determine number of data points
    N = int( tmax/dt + 1 )

    #Initialize output vectors for t,I,S,D
    t = [0]*N
    S = [S0]*N
    I = [I0]*N
    D = [0]*N
	
    #Loop through vectors to compute t, I, S, D.
    #using the Euler first order differential equation method
    for n in range(1,N):
        t[n] = t[n-1] + dt
        I[n] = I[n-1] + dt*( beta*S[n-1]*I[n-1] - alpha*I[n-1] )
        S[n] = S[n-1] - dt*beta*S[n-1]*I[n-1]
        D[n] = D[n-1] + dt*alpha*I[n-1]
    
    #Set return values of function
    return t, I, S, D

#

#Line of best fit function yfit = m*x +c, with product moment correlation
#coefficient r
def bestfitc(x,y):
    
    #Find any x or y values that are NaN or Inf
    test = Npy.logical_or( Npy.isnan(Npy.abs(x)),Npy.isnan(Npy.abs(y)) )
    test = Npy.logical_or( test, Npy.isinf(Npy.abs(x)) )
    test = Npy.logical_or( test, Npy.isinf(Npy.abs(y)) )
    i = Npy.nonzero(test)
    x = Npy.delete(x,i,None)
    y = Npy.delete(y,i,None)

    #Compute line of best fit
    xbar = Npy.mean(x)
    ybar = Npy.mean(y)
    xybar = Npy.mean(x*y)
    Vx = Npy.mean(x**2 ) - xbar**2
    Vy = Npy.mean(y**2 ) - ybar**2
    COVxy = Npy.mean(x*y) - xbar*ybar
    m = COVxy/Vx
    c = ybar - m*xbar
    r = COVxy/Npy.sqrt( Vx*Vy )
    yfit = m*x + c
    xfit = x

    #Return outputs
    return yfit,xfit,r,ybar,xbar,m,c

#

#Line of best fit function yfit = m*x, with product moment correlation
#coefficient r
def bestfit(x,y):
    
    #Find any x or y values that are NaN or Inf
    test = Npy.logical_or( Npy.isnan(Npy.abs(x)),Npy.isnan(Npy.abs(y)) )
    test = Npy.logical_or( test, Npy.isinf(Npy.abs(x)) )
    test = Npy.logical_or( test, Npy.isinf(Npy.abs(y)) )
    i = Npy.nonzero(test)
    x = Npy.delete(x,i,None)
    y = Npy.delete(y,i,None)

    #Compute line of best fit
    xybar = Npy.mean(x*y)
    xxbar = Npy.mean(x**2 )
    yybar = Npy.mean(y**2 )
    m = xybar/xxbar
    r = xybar/( xxbar*yybar )
    yfit = m*x
    xfit = x

    #Return outputs
    return yfit,xfit,r,m

#

#Display greeting message
print(" \n EYAM PLAGUE EPIDEMIOLOGICAL MODEL ")
print(" Dr Andrew French. February 2018 " )
print(" \n Plotting Mompesson's Eyam data ... " )

# Mompesson Eyam data
t_data = [0,0.51,1.02,1.53,2.04,2.55,3.57]
I_data = [14.5,22,29,21,8,8,0]
S_data = [235,201,153.5,121,108,97,83]

#Determine derived parameters
D_data = [0]*len(t_data)
x = [0]*len(t_data)
y = [0]*len(t_data)
for i in range(0,len(t_data)):
  D_data[i] = I_data[0] + S_data[0] - I_data[i] - S_data[i]
  x[i] = math.log( S_data[0]/S_data[i] )
  y[i] = I_data[0] + S_data[0] - I_data[i] - S_data[i]

#Plot Mompesson's Eyam data
fig = plt.figure(figsize=(10, 8))
plt.plot( t_data,I_data,'-+',label = "I" , color = [1,0,0], markersize = msize )
plt.plot( t_data,S_data,'-+',label = "S", color = [0,1,0], markersize = msize  )
plt.plot( t_data,D_data,'-+',label = "D", color = [0,0,1], markersize = msize  )
plt.xlabel( "t /months", size=fsize );
plt.ylabel( "Eyam population", size=fsize );
plt.xticks( fontsize = fsize )
plt.yticks( fontsize = fsize )
plt.title( "Mompesson's Eyam data", size=fsize )
plt.legend( fontsize = fsize,loc = "upper right" )
plt.xlim( 0, tmax  )
plt.grid(True)

#Export PNG file
fig.savefig('Mompesson data.png', dpi=300)

#

#Determining Eyam model parameter ratio alpha/beta
#from I and S data
print("\n Determining Eyam model parameter ratio alpha/beta from I and S data ..." )

#Determine line of best fit through x,y values formed from Eyam
#I and S data
yfit,xfit,r,alpha_over_beta = bestfit(x,y)

#Plot best fit to Eyam data to determine alpha/beta ratio
fig = plt.figure(figsize=(10, 8))
plt.plot( x,y,'+',label = "Eyam data" , color = [1,0,0], markersize = msize  )
plt.plot( xfit,yfit,'-',label = "linear fit" , color = [0,0,0] )
plt.xlabel( "ln( S0/S )", size=fsize );
plt.ylabel( "I0 + S0 - I - S", size=fsize );
plt.xticks( fontsize = fsize )
plt.yticks( fontsize = fsize )
plt.title( "alpha/beta = " + str(Npy.around(alpha_over_beta,3)), size=fsize )
plt.legend( fontsize = fsize,loc = "upper left" )
plt.xlim( 0, 1.1 )
plt.ylim( 0, 175 )
plt.grid(True)

#Export PNG file
fig.savefig('Mompesson alpha over beta calc.png', dpi=300)

#

#Eyam model
print( "\n Running Eyam model for a range of alpha values ..." )

#Compute number of data points i.e. xmax = x0 + (N-1)*dx


#Run model in a loop and determine alpha parameter which minimizes the sum
#of square differences (SSD) from the Eyam data
alpha = Npy.linspace(alphamin,alphamax,num_alpha)
SSD = [0]*num_alpha
for i in range(0,num_alpha):

    #Display message to console window
    print( " Alpha iteration number = " + str(i) + " of " + str(num_alpha) )
           
    #Run Eyam model
    beta = alpha[i]/alpha_over_beta
    t, I, S, D = eyam_model( dt, I_data[0], S_data[0], alpha[i], beta, tmax )

    #Interpolate at Eyam data times
    Ii = Npy.interp( t_data, t, I )
    Si = Npy.interp( t_data, t, S )
    Di = Npy.interp( t_data, t, D )

    #Determine sum of square differences between model and data
    SSD[i] = Npy.sum( (I_data - Ii)**2 + (S_data - Si)**2 + (D_data - Di)**2 )

#Find alpha parameter which corresponds to minimum SSD
amin = alpha[ SSD==Npy.min(SSD) ]
amin = amin[0]

#Plot SSD vs alpha
fig = plt.figure(figsize=(10, 8))
plt.plot( alpha,SSD,'-',label = "SSD" , color = [1,0,0] )
plt.plot( amin,Npy.min(SSD),'+',label = "SSD min" , color = [0,1,0], markersize = msize )
plt.xlabel( "alpha", size=fsize );
plt.ylabel( "Sum of square differences", size=fsize );
plt.xticks( fontsize = fsize )
plt.yticks( fontsize = fsize )
plt.title( "alpha = " + str(Npy.around(amin,3)) + " for minimum SSD ", size=fsize )
plt.legend( fontsize = fsize,loc = "upper left" )
plt.xlim( alphamin, alphamax )
plt.grid(True)

#Export PNG file
fig.savefig('SSD vs alpha.png', dpi=300)

#

#Eyam model, with alpha corresponding to minimum FSD, with Mompesson data overlay
print("\n Computing optimal Eyam model and plotting it with Mompesson data overlay ")

#Solve Eyam model
alpha = amin
beta = alpha/alpha_over_beta
t, I, S, D = eyam_model( dt, I_data[0], S_data[0], alpha, beta, tmax )

#Plot Eyam model and overlay Mompesson data
fig = plt.figure(figsize=(10, 8))
plt.plot( t,I,'-',label = "I" , color = [1,0,0] )
plt.plot( t,S,'-',label = "S", color = [0,1,0] )
plt.plot( t,D,'-',label = "D", color = [0,0,1] )
plt.plot( t_data,I_data,'+',label = "I data" , color = [1,0,0], markersize = msize  )
plt.plot( t_data,S_data,'+',label = "S data", color = [0,1,0], markersize = msize  )
plt.plot( t_data,D_data,'+',label = "D data", color = [0,0,1], markersize = msize  )
plt.xlabel( "t /months", size=fsize );
plt.ylabel( "Eyam population", size=fsize );
plt.xticks( fontsize = fsize )
plt.yticks( fontsize = fsize )
plt.title( "Eyam model: alpha = " + str(Npy.around(alpha,3))+", beta = " +
           str(Npy.around(beta,3)), size=fsize )
plt.legend( fontsize = fsize,loc = "upper left" )
plt.xlim( 0, tmax )
plt.grid(True)

#Export PNG file
fig.savefig('Eyam model.png', dpi=300)

#

#Wait for user input to close terminal window
input("\n Press return key to close ...")

#End of code
